#!/bin/bash

# Copyright (c) 2019-2021, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This is a script to parse control file, such as "upgradetasklist.txt" to
# update partitions in partition-based OTA

# Each line in control file includes 5 elements which are:
# <task_stage>:<action>:<device>:<partition>:<index file>

# Where
#
#  <task_stage>: Update_A, Update_B, Update_B1, Update_B2, ...
#
#  <action>: Write, Write_BCT1, Write_BCT2, Write_APP, Write_APP_b, Clean_BCT1, Read, Set_Boot_Normal_Mode, Set_Boot_Recovery_Mode, Set_Boot_Slot_A, Set_Boot_Slot_B
#
#  <device>: sdmmc_boot, sdmmc_user
#
#  <partition>: BCT, mb1, mb1_b, MB1_BCT, MB1_BCT_b, dram-ecc-fw, badpage-fw, badpage-fw_b, spe-fw, spe-fw_b, mb2, mb2_b, mts-preboot, mts-preboot_b, SMD, SMD_b, secondary_gpt, master_boot_record, primary_gpt, APP, mts-bootpack, mts-bootpack_b, cpu-bootloader, cpu-bootloader_b, bootloader-dtb, bootloader-dtb_b, secure-os, secure-os_b, eks, adsp-fw, adsp-fw_b, bpmp-fw, bpmp-fw_b, bpmp-fw-dtb, bpmp-fw-dtb_b, sce-fw, sce-fw_b, sc7, sc7_b, FBNAME, BMP, BMP_b, recovery, recovery-dtb, kernel-bootctrl, kernel-bootctrl_b, kernel, kernel_b, kernel-dtb, kernel-dtb_b, secondary_gpt
#
#  <index file>: <directory of images>/<index file name>, such as R32-ToT/flash.idx, R32_1_2-R32i/flash.idx

ACTIVE_INDEX_FILE=""
ACTIVE_INDEX_ARRAY=
APP_PARTITION=""
APP_B_PARTITION=""
BCT_BLOCK_SIZE=16384
BOOT_DEV_SECTOR_SIZE=
COMMON_IMAGES_DIR=
CURRENT_STAGE=""
K_BYTES=1024
ROOTFS_UPDATE_SUCCESS_FLAG="/tmp/rootfs_update_success"
SDMMC_BOOT0_SIZE=4194304
SPI_BOOT_DEVICE="/dev/mtd0"
OTA_WORK_DIR=
VERIFY_WRITE=1

usage()
{
	echo "Usage: $0 [options] <control file> "
	cat << EOF
    options:
	-s ------------------------- skip verifying after writing data. Default: disabled
	-h ------------------------- print this message
EOF
	exit 1
}

# Verify sha1 checksum for image
# @file_image: file for caculating check sum
# @sha1chksum: sha1 check sum
sha1_verify ()
{
	local file_image=$1
	local sha1_chksum=$2

	if [ -z "${sha1_chksum}" ]; then
		echo "Passed-in sha1 checksum is NULL"
		return 1
	fi

	if [ ! -f "${file_image}" ]; then
		echo "$file_image is not found !!!"
		return 1
	fi

	local sha1_chksum_gen=
	sha1_chksum_gen="$(sha1sum "${file_image}" | cut -d\  -f 1)"
	if [ "${sha1_chksum_gen}" = "${sha1_chksum}" ]; then
		echo "Sha1 checksum matched for ${file_image}"
		return 0
	else
		echo "Sha1 checksum does not match (${sha1_chksum_gen} != ${sha1_chksum}) for ${file_image}"
		return 1
	fi
}

rw_part_opt()
{
	local infile=$1
	local outfile=$2
	local inoffset=$3
	local outoffset=$4
	local size=$5

	if [ ! -e "${infile}" ]; then
		echo "Input file ${infile} is not found"
		return 1
	fi

	if [ "${size}" == "0" ]; then
		echo "The size of bytes to be read is ${size}"
		return 1
	fi

	local inoffset_align_K=$((inoffset % K_BYTES))
	local outoffset_align_K=$((outoffset % K_BYTES))
	if [ ${inoffset_align_K} -ne 0 ] || [ ${outoffset_align_K} -ne 0 ]; then
		echo "Offset is not aligned to K Bytes, no optimization is applied"
		echo "dd if=${infile} of=${outfile} bs=1 skip=${inoffset} seek=${outoffset} count=${size}"
		dd if="${infile}" of="${outfile}" bs=1 skip="${inoffset}" seek="${outoffset}" count="${size}"
		return 0
	fi

	local block=$((size / K_BYTES))
	local remainder=$((size % K_BYTES))
	local inoffset_blk=$((inoffset / K_BYTES))
	local outoffset_blk=$((outoffset / K_BYTES))

	echo "${size} bytes from ${infile} to ${outfile}: 1KB block=${block} remainder=${remainder}"

	if [ ${block} -gt 0 ]; then
		echo "dd if=${infile} of=${outfile} bs=1K skip=${inoffset_blk} seek=${outoffset_blk} count=${block}"
		dd if="${infile}" of="${outfile}" bs=1K skip="${inoffset_blk}" seek="${outoffset_blk}" count="${block}" conv=notrunc
		sync
	fi
	if [ ${remainder} -gt 0 ]; then
		local block_size=$((block * K_BYTES))
		local outoffset_rem=$((outoffset + block_size))
		local inoffset_rem=$((inoffset + block_size))
		echo "dd if=${infile} of=${outfile} bs=1 skip=${inoffset_rem} seek=${outoffset_rem} count=${remainder}"
		dd if="${infile}" of="${outfile}" bs=1 skip="${inoffset_rem}" seek="${outoffset_rem}" count="${remainder}" conv=notrunc
		sync
	fi
	return 0
}

flash_partition()
{
	local file_name=$1
	local part_name=$2
	local start_offset=$3
	local file_size=$4
	local attributes=$5
	local sha1_chksum=$6
	local sdmmc_device=$7
	local tmp_file=/tmp/tmp.img
	local active_dir=
	local file_image=
	local sha1_chksum_gen=
	local res=0
	local tmp_size=0

	active_dir="$(dirname "${ACTIVE_INDEX_FILE}")"
	file_image="${active_dir}/${file_name}"

	if [ ! -f "${file_image}" ]; then
		file_image=${COMMON_IMAGES_DIR}/${file_name}
	fi

	# skip verifying SMD/SMD_b/kernel-bootctrl/kernel-bootctrl_b
	# paritions as the sha1 cheksum in the index file is generated
	# from the dummy image file other than the exact image file
	if [ "${part_name}" != "SMD" ] \
		&& [ "${part_name}" != "SMD_b" ] \
		&& [ "${part_name}" != "kernel-bootctrl" ] \
		&& [ "${part_name}" != "kernel-bootctrl_b" ]; then
		if ! sha1_verify "${file_image}" "${sha1_chksum}"; then
			return 1
		fi
	else
		sha1_chksum="$(sha1sum "${file_image}" | cut -d\  -f 1)"
		if [ "${sha1_chksum}" = "" ];then
			echo "Failed to generate sha1sum for file ${file_image}"
			return 1
		fi

	fi

	# verify whether this partition has been writen
	rm "${tmp_file}"
	echo "read-back partition ${part_name} to check whether to update it"
	if [ "${sdmmc_device}" != "" ]; then
		if ! rw_part_opt "${sdmmc_device}" "${tmp_file}" "${start_offset}" 0 "${file_size}"; then
			echo "Failed to read ${file_size} bytes from ${sdmmc_device}:${start_offset} to ${tmp_file}"
			return 1
		fi
	else
		tmp_size=$((SDMMC_BOOT0_SIZE - start_offset))
		dd if=/dev/mmcblk0boot0 of="${tmp_file}" skip="${start_offset}" bs=1 count="${tmp_size}"
		tmp_size=$((file_size - tmp_size))
		dd if=/dev/mmcblk0boot1 bs=1 count="${tmp_size}" >>"${tmp_file}"
	fi
	sync
	sha1_chksum_gen="$(sha1sum "${tmp_file}" | cut -d\  -f 1)"
	if [ "${sha1_chksum_gen}" = "${sha1_chksum}" ]; then
		echo "Partition ${part_name} has been updated, skip writing"
		return 0
	fi

	# write image
	echo "write ${file_image} into partition ${part_name}"
	if [ "${sdmmc_device}" != "" ]; then
		if ! rw_part_opt "${file_image}" "${sdmmc_device}" 0 "${start_offset}" "${file_size}"; then
			echo "Failed to write ${file_size} bytes from ${file_image} to ${sdmmc_device}:${start_offset}"
			return 1
		fi
	else
		tmp_size=$((SDMMC_BOOT0_SIZE - start_offset))
		echo "dd if=${file_image} of=${sdmmc_device} seek=${start_offset} bs=1 count=${tmp_size} conv=notrunc"
		dd if="${file_image}" of=/dev/mmcblk0boot0 seek="${start_offset}" bs=1 count="${tmp_size}" conv=notrunc
		dd if="${file_image}" of=/dev/mmcblk0boot1 bs=1 skip="${tmp_size}" conv=notrunc
	fi
	sync

	if [ ${VERIFY_WRITE} -eq 1 ]; then
		rm "${tmp_file}"
		# verify writing
		if [ "${sdmmc_device}" != "" ]; then
			if ! rw_part_opt "${sdmmc_device}" "${tmp_file}" "${start_offset}" 0 "${file_size}"; then
				echo "Failed to read ${file_size} bytes from ${sdmmc_device}:${start_offset} to ${tmp_file}"
				return 1
			fi
		else
			tmp_size=$((SDMMC_BOOT0_SIZE - start_offset))
			echo "dd if=/dev/mmcblk0boot0 of=${tmp_file} skip=${start_offset} bs=1 count=${tmp_size}"
			dd if=/dev/mmcblk0boot0 of="${tmp_file}" skip="${start_offset}" bs=1 count="${tmp_size}"
			tmp_size=$((file_size - tmp_size))
			echo "dd if=/dev/mmcblk0boot1 bs=1 count=${tmp_size} >>${tmp_file}"
			dd if=/dev/mmcblk0boot1 bs=1 count="${tmp_size}" >>"${tmp_file}"
		fi

		# For SMD/SMD_B and kernel-bootctrl/kernel-bootctrl_b, the
		# sha1 chksum needs to be re-generated from the exact image file
		if [ "${part_name}" = "SMD" ] \
			|| [ "${part_name}" = "SMD_b" ] \
			|| [ "${part_name}" = "kernel-bootctrl" ] \
			|| [ "${part_name}" = "kernel-bootctrl_b" ]; then
			echo "Re-generate sha1sum for the image ${file_image}"
			sha1_chksum="$(sha1sum "${file_image}" | cut -d\  -f 1)"
		fi

		sha1_verify "${tmp_file}" "${sha1_chksum}"
	fi

	return $?
}

flash_sdmmc_boot_partition()
{
	local start_offset=$3
	local file_size=$4
	local sdmmc_device=
	local args=("$@")
	local end_offset=

	start_offset=$((start_offset))
	file_size=$((file_size))
	end_offset=$((start_offset + file_size))
	if [ ${start_offset} -ge ${SDMMC_BOOT0_SIZE} ]; then
		sdmmc_device=/dev/mmcblk0boot1
		start_offset=$((start_offset - SDMMC_BOOT0_SIZE))
		args[2]=${start_offset}
	elif [ ${end_offset} -lt ${SDMMC_BOOT0_SIZE} ]; then
		sdmmc_device=/dev/mmcblk0boot0
	else
		# partition cross over mmcblk0boot0 and mmcblk0boot1 and
		# it should be handled in special way
		sdmmc_device=""
	fi

	flash_partition ${args[@]} "${sdmmc_device}"
	return $?
}

flash_sdmmc_user_partition()
{
	local sdmmc_device=/dev/mmcblk0

	flash_partition $@ "${sdmmc_device}"
	return $?
}

read_partition ()
{
	local file_name=$1
	local part_name=$2
	local start_offset=$3
	local file_size=$4
	local attributes=$5
	local sha1_chksum=$6
	local sdmmc_device=$7
	local active_dir=
	local file_image=
	local tmp_size=0

	active_dir="$(dirname "${ACTIVE_INDEX_FILE}")"
	file_image="${active_dir}/${file_name}"

	# read partition
	if [ "${sdmmc_device}" != "" ]; then
		echo "dd if=${sdmmc_device}} of=${file_image}_read skip=${start_offset} \
			bs=1 count=${file_size} conv=notrunc"
		if ! rw_part_opt "${sdmmc_device}" "${file_image}_read" "${start_offset}" 0 "${file_size}"; then
			echo "Failed to read ${file_size} bytes from ${sdmmc_device}:${start_offset} to ${file_image}_read"
			return 1
		fi
	else
		tmp_size=$((SDMMC_BOOT0_SIZE - start_offset))
		echo "dd if=/dev/mmcblk0boot0 of=${file_image} skip=${start_offset} bs=1 count=${tmp_size}"
		dd if=/dev/mmcblk0boot0 of="${file_image}" skip="${start_offset}" bs=1 count="${tmp_size}"
		tmp_size=$((file_size - tmp_size))
		echo "dd if=/dev/mmcblk0boot1 bs=1 count=${tmp_size} >>${file_image}"
		dd if=/dev/mmcblk0boot1 bs=1 count="${tmp_size}" >>"${file_image}"
	fi
	sync
	return 0
}

read_sdmmc_boot_partition()
{
	local start_offset=$3
	local file_size=$4
	local sdmmc_device=
	local args=("$@")
	local end_offset=

	start_offset=$((start_offset))
	file_size=$((file_size))
	end_offset=$((start_offset + file_size))
	if [ ${start_offset} -ge ${SDMMC_BOOT0_SIZE} ]; then
		sdmmc_device=/dev/mmcblk0boot1
		start_offset=$((start_offset - SDMMC_BOOT0_SIZE))
		args[2]=${start_offset}
	elif [ ${end_offset} -lt ${SDMMC_BOOT0_SIZE} ]; then
		sdmmc_device=/dev/mmcblk0boot0
	else
		# partition cross over mmcblk0boot0 and mmcblk0boot1 and
		# it should be handled in special way
		sdmmc_device=""
	fi

	read_partition ${args[@]}  "${sdmmc_device}"
	return $?
}

read_sdmmc_user_partition()
{
	local sdmmc_device=/dev/mmcblk0

	read_partition $@ "${sdmmc_device}"
	return $?
}

do_read()
{
	local item=$1
	local device_type=
	local part_name=
	local file_name=
	local start_offset=
	local file_size=
	local attributes=
	local sha1_chksum=
	local res=0

	device_type="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 1)"
	part_name="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 3)"
	file_name="$(echo "$item" | cut -d, -f 5 | sed 's/^ //g' -)"
	start_offset="$(echo "$item" | cut -d, -f 3 | sed 's/^ //g' -)"
	file_size="$(echo "$item" | cut -d, -f 6 | sed 's/^ //g' -)"
	attributes="$(echo "$item" | cut -d, -f 7 | sed 's/^ //g' -)"
	sha1_chksum="$(echo "$item" | cut -d, -f 8 | sed 's/^ //g' -)"

	if [ "${file_name}" = "" ]; then
		echo "Error: file name is not specified for reading ${part_name} partition"
		return 1
	fi

	echo "Reading ${part_name} partition into ${file_name}"
	# device_type 0 means "sdmmc_boot", 1 means "sdmmc_user". Other values are invalid.
	if [ "${device_type}" = "0" ]; then
		res=$(read_sdmmc_boot_partition "${file_name}" "${part_name}" \
			"${start_offset}" "${file_size}" "${attributes}" "${sha1_chksum}")
	elif [ "${device_type}" = "1" ]; then
		res=$(read_sdmmc_user_partition "${file_name}" "${part_name}" \
			"${start_offset}" "${file_size}" "${attributes}" "${sha1_chksum}")
	else
		echo "Error: invalid device type ${device_type}"
		res=1
	fi
	echo "Reading ${part_name} partition done"
	return ${res}
}

do_write()
{
	local item=$1
	local action=$2
	local sha1_chksum=
	local device_type=
	local part_name=
	local file_name=
	local start_offset=
	local file_size=
	local attributes=
	local sha1_chksum=
	local res=0

	sha1_chksum="$(echo "$item" | cut -d, -f 8 | sed 's/^ //g' -)"
	device_type="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 1)"
	part_name="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 3)"
	file_name="$(echo "$item" | cut -d, -f 5 | sed 's/^ //g' -)"
	start_offset="$(echo "$item" | cut -d, -f 3 | sed 's/^ //g' -)"
	file_size="$(echo "$item" | cut -d, -f 6 | sed 's/^ //g' -)"
	attributes="$(echo "$item" | cut -d, -f 7 | sed 's/^ //g' -)"
	sha1_chksum="$(echo "$item" | cut -d, -f 8 | sed 's/^ //g' -)"

	if [ "${file_name}" = "" ]; then
		echo "Warning: skip writing ${part_name} partition as no image is specified"
		return 0
	fi

	echo "Writing ${part_name} partition with ${file_name}"
	# device_type 0 means "sdmmc_boot", 1 means "sdmmc_user". Other values are invalid.
	if [ "${device_type}" = "0" ]; then
		if [ "${action}" == "Write_BCT2" ]; then
			start_offset=$((start_offset + BCT_BLOCK_SIZE))
		fi
		flash_sdmmc_boot_partition "${file_name}" "${part_name}" \
			${start_offset} "${file_size}" "${attributes}" "${sha1_chksum}"
		res=$?
		if [ ${res} -ne 0 ]; then
			return ${res}
		fi

		# need to write block0/slot1 as well for BCT partition
		if [ "${action}" == "Write_BCT1" ];then
			# align to sector when writing block0/slot1
			local remainder=$(((start_offset + file_size) % BOOT_DEV_SECTOR_SIZE))
			if [ ${remainder} -eq 0 ];then
				start_offset=$((start_offset + file_size))
			else
				start_offset=$((start_offset + file_size + BOOT_DEV_SECTOR_SIZE - remainder))
			fi
			flash_sdmmc_boot_partition "${file_name}" "${part_name}" \
				${start_offset} "${file_size}" "${attributes}" "${sha1_chksum}"
			res=$?
		fi
	elif [ "${device_type}" = "1" ]; then
		flash_sdmmc_user_partition "${file_name}" "${part_name}" \
			${start_offset} "${file_size}" "${attributes}" "${sha1_chksum}"
		res=$?
	else
		echo "Error: invalid device type ${device_type}"
		return 1
	fi
	echo "Writing ${part_name} partition done"
	return ${res}
}

write_APP()
{
	local file_image="${1}"
	local app_partition="${2}"

	# Set rootfs updater and update rootfs partition with it
	local rootfs_updater=
	if ! set_rootfs_updater "${OTA_WORK_DIR}" rootfs_updater; then
		ota_log "Failed to run \"set_rootfs_updater ${OTA_WORK_DIR} rootfs_updater\""
		return 1
	fi
	echo "eval ${rootfs_updater} -p ${app_partition} -d ${OTA_WORK_DIR} ${file_image}"
	eval "${rootfs_updater}" -p "${app_partition}" -d "${OTA_WORK_DIR}" "${file_image}"
	if [ ! -e "${ROOTFS_UPDATE_SUCCESS_FLAG}" ]; then
		echo "Failed to run \"eval ${rootfs_updater} -p ${app_partition} -d ${OTA_WORK_DIR} ${file_image}\""
		return 1
	fi
	return 0
}

do_write_APP()
{
	local item=$1
	local action=$2
	local partition_name="$3"
	local file_image=${COMMON_IMAGES_DIR}/system.img
	local sha1_file=${COMMON_IMAGES_DIR}/system.img.sha1sum
	local sha1_chksum=

	if [ ! -f "${file_image}" ]; then
		echo "APP image ${file_image} is not found !!!"
		return 1
	fi

	if [ ! -f "${sha1_file}" ]; then
		echo "Sha1 checksum file ${sha1_file} is not found !!!"
		return 1
	fi

	local app_partition=
	if [ "${partition_name}" == "APP" ]; then
		app_partition="${APP_PARTITION}"
	else
		app_partition="${APP_B_PARTITION}"
	fi
	if [ ! -e "${app_partition}" ]; then
		echo "APP paritiion ${app_partition} is not found !!!"
		return 1
	fi

	# verify sha1 checksum
	sha1_chksum="$(cat "${sha1_file}")"
	if ! sha1_verify "${file_image}" "${sha1_chksum}"; then
		return 1
	fi

	if ! write_APP "${file_image}" "${app_partition}"; then
		echo "Failed to run \"Write_APP ${file_image} ${app_partition}\""
		return 1
	fi
	return 0
}

do_set_boot_slot()
{
	local item=$1
	local bootslot=$2
	local file_name=
	local active_dir=
	local file_image=
	file_name="$(echo "$item" | cut -d, -f 5 | sed 's/^ //g' -)"
	active_dir="$(dirname "${ACTIVE_INDEX_FILE}")"
	file_image="${active_dir}/${file_name}"

	if [ ! -f "${file_image}.${bootslot}" ]; then
		echo "${file_image}.${bootslot} is not found !!!"
		return 1
	fi
	echo "cp -f \"${file_image}.${bootslot}\" \"${file_image}\""
	cp -f "${file_image}.${bootslot}" "${file_image}"

	do_write "${item}" "Write"
	return $?
}

do_set_boot_mode()
{
	local item=$1
	local bootmode=$2
	local file_name=
	local active_dir=
	local file_image=
	file_name="$(echo "$item" | cut -d, -f 5 | sed 's/^ //g' -)"
	active_dir="$(dirname "${ACTIVE_INDEX_FILE}")"
	file_image="${active_dir}/${file_name}"

	if [ ! -f "${file_image}.${bootmode}" ]; then
		echo "${file_image}.${bootmode} is not found !!!"
		return 1
	fi
	echo "cp -f \"${file_image}.${bootmode}\" \"${file_image}\""
	cp -f "${file_image}.${bootmode}" "${file_image}"

	do_write "${item}" "Write"
	return $?
}

do_clean_BCT1()
{
	local item=$1
	local sdmmc_device=/dev/mmcblk0boot0
	local start_offset=
	local file_size=
	start_offset=$(echo "$item" | cut -d, -f 3 | sed 's/^ //g' -)
	file_size=$(echo "$item" | cut -d, -f 6 | sed 's/^ //g' -)

	# BCT1(block 0) has 2 copies: slot 0 and slot 1
	# align to sector size
	local remainder=$((file_size % BOOT_DEV_SECTOR_SIZE))
	local write_size=
	if [ ${remainder} -eq 0 ];then
		write_size=$((file_size * 2))
	else
		write_size=$((file_size + BOOT_DEV_SECTOR_SIZE - remainder))
		write_size=$((write_size * 2))
	fi

	echo "Clean the BCT1(block 0) slot 0 and slot 1"
	echo "dd if=/dev/zero of=${sdmmc_device} bs=1 seek=${start_offset} count=${write_size}"
	dd if=/dev/zero of="${sdmmc_device}" bs=1 seek="${start_offset}" count="${write_size}"
	res=$?
	sync

	return ${res}
}

do_action()
{
	local action=$1
	local item=$2
	local res=0

	case ${action} in
		Write|Write_BCT1|Write_BCT2)
			echo "Do write action"
			do_write "${item}" "${action}"
			return $?
			;;
		Write_APP)
			echo "Do write APP action"
			do_write_APP "${item}" "${action}" "APP"
			return $?
			;;
		Write_APP_b)
			echo "Do write APP_b action"
			do_write_APP "${item}" "${action}" "APP_b"
			return $?
			;;
		Read)
			echo "Do read action"
			do_read "${item}"
			return $?
			;;
		Set_Boot_Recovery_Mode)
			echo "Do set boot to recovery mode"
			do_set_boot_mode "${item}" "update"
			return $?
			;;
		Set_Boot_Normal_Mode)
			echo "Do set boot to normal mode"
			do_set_boot_mode "${item}" "normal"
			return $?
			;;
		Set_Boot_Slot_A)
			echo "Do set boot to slot A"
			do_set_boot_slot "${item}" "A"
			return $?
			;;
		Set_Boot_Slot_B)
			echo "Do set boot to slot B"
			do_set_boot_slot "${item}" "B"
			return $?
			;;
		Clean_BCT1)
			echo "Do clean BCT1"
			do_clean_BCT1 "${item}"
			return $?
			;;
		*)
			echo "Invalid action ${action}"
			return 1
	esac
}

run_task()
{
	local action=$1
	local device=$2
	local partition=$3
	local index_file=$4
	local device_type=
	local res=0

	if [ "${index_file}" != "${ACTIVE_INDEX_FILE}" ]; then
		readarray ACTIVE_INDEX_ARRAY < "${index_file}"
		ACTIVE_INDEX_FILE="${index_file}"
	fi
	echo "Active index file is ${ACTIVE_INDEX_FILE}"

	local lines_num=${#ACTIVE_INDEX_ARRAY[@]}
	echo "Number of lines is $lines_num"

	local max_index=$((lines_num - 1))
	echo "max_index=${max_index}"

	local item=
	local part_name=
	for i in $(seq 0 ${max_index})
	do
		item=${ACTIVE_INDEX_ARRAY[$i]}

		part_name="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 3)"
		if [ "${part_name}" != "${partition}" ]; then
			continue
		fi

		device_type="$(echo "$item" | cut -d, -f 2 | sed 's/^ //g' - | cut -d: -f 1)"
		if { [ "${device_type}" = "0" ] && [ "${device}" = "sdmmc_boot" ]; } \
			|| { [ "${device_type}" = "1" ] && [ "${device}" = "sdmmc_user" ]; }; then
			echo -n "action=${action} item=${item}"
			do_action "${action}" "${item}"
			return $?
		fi
	done
}


if [ $# -lt 1 ]; then
	usage
fi

nargs=$#;
control_file=${!nargs};

if [ ! -f "${control_file}" ]; then
	echo "Specified control file ${control_file} is not found"
	usage
fi

opstr+="h:s:-:";
while getopts "${opstr}" OPTION; do
	case $OPTION in
	h) usage; ;;
	s) VERIFY_WRITE=0; echo "VERIFY_WRITE=0";;
	*) usage; ;;
	esac;
done

APP_PARTITION="$(blkid | grep -m 1 "PARTLABEL=\"APP\"" | cut -d: -f 1)"
APP_B_PARTITION="$(blkid | grep -m 1 "PARTLABEL=\"APP_b\"" | cut -d: -f 1)"
OTA_WORK_DIR="$(cd "$(dirname "${control_file}")" && pwd)"
COMMON_IMAGES_DIR="${OTA_WORK_DIR}"

echo "Control file is ${control_file}, under ${OTA_WORK_DIR}"

readarray controlarray < "${control_file}"

lines_num=${#controlarray[@]}
echo "Number of lines is $lines_num"

max_index=$((lines_num - 1))
echo "max_index=${max_index}"

# make mmcblk0bootX writable
echo 0 >/sys/block/mmcblk0boot0/force_ro
echo 0 >/sys/block/mmcblk0boot1/force_ro

# get sector size of boot devie and user device
if [ -e "${SPI_BOOT_DEVICE}" ];then
	BOOT_DEV_SECTOR_SIZE=$(cat /sys/block/mtdblock0/queue/hw_sector_size)
else
	BOOT_DEV_SECTOR_SIZE=$(cat /sys/block/mmcblk0boot0/queue/hw_sector_size)
	SDMMC_BOOT0_SIZE=$(cat /sys/block/mmcblk0boot0/size)
	SDMMC_BOOT0_SIZE=$((SDMMC_BOOT0_SIZE * 1024 / 2))
fi

set +e
for i in $(seq 0 ${max_index})
do
	item="${controlarray[$i]}"
	echo -n "line[$i]: ${item}"

	# skip empty line
	if [ "${#item}" = "1" ]; then
		continue
	fi

	# skip comment line
	if [[ "${item}" =~ ^# ]]; then
		continue
	fi

	# skip invalid line
	if [[ ! "${item}" =~ Write|Read|Set_Boot|Clean ]]; then
		continue
	fi

	stage="$(echo "${item}" | cut -d: -f 1)"
	action="$(echo "${item}" | cut -d: -f 2)"
	device="$(echo "${item}" | cut -d: -f 3)"
	partition="$(echo "${item}" | cut -d: -f 4)"
	index_file="$(echo "${item}" | cut -d: -f 5)"
	index_file="${OTA_WORK_DIR}/${index_file}"

	echo "stage=${stage} action=${action} device=${device} partition=${partition} index_file=${index_file}"

	if [ ! -f "${index_file}" ]; then
		echo "Index file ${index_file} is not found"
		continue
	fi

	if [ "${stage}" != "${CURRENT_STAGE}" ]; then
		CURRENT_STAGE=${stage}
		echo "Task stage ${CURRENT_STAGE} starts"
	fi
	if ! run_task "${action}" "${device}" "${partition}" "${index_file}"; then
		echo "Failed to run_task ${action} ${device} ${partition} ${index_file}"
		exit 1
	fi
done

# set mmcblk0bootX readonly
echo 1 >/sys/block/mmcblk0boot0/force_ro
echo 1 >/sys/block/mmcblk0boot1/force_ro


# mark OTA is success
echo 1 >/tmp/ota_success
